'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});

var _create = require('babel-runtime/core-js/object/create');

var _create2 = _interopRequireDefault(_create);

exports.default = memoize;

var _lodash = require('lodash');

var _lodash2 = _interopRequireDefault(_lodash);

var _hiddenData = require('../hiddenData');

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

function memoize(target, property, descriptor) {
  var cacheProp = 'memoized' + _lodash2.default.upperFirst(property);
  var impl = descriptor.value;

  if (impl.length === 0) {
    descriptor.value = createSingleValueMemoizedFunc(impl, cacheProp);
  } else {
    descriptor.value = createMultiValueMemoizedFunc(impl, cacheProp);
  }
}

function createSingleValueMemoizedFunc(impl, cacheProp) {
  var get = (0, _hiddenData.createHiddenDataGetter)(cacheProp);
  var set = (0, _hiddenData.createHiddenDataSetter)(cacheProp);

  return function decorator$memoize() {
    var value = get(this);

    if (value === undefined) {
      value = impl.call(this);
      set(this, value);
    }

    return value;
  };
}

function createMultiValueMemoizedFunc(impl, cacheProp) {
  var get = (0, _hiddenData.createHiddenDataGetter)(cacheProp);
  var set = (0, _hiddenData.createHiddenDataSetter)(cacheProp);

  return function decorator$memoize(input) {
    var cache = get(this);

    if (cache === undefined) {
      cache = (0, _create2.default)(null);
      set(this, cache);
    }

    if (input in cache) {
      return cache[input];
    } else {
      var value = impl.call(this, input);
      cache[input] = value;
      return value;
    }
  };
}